#define TOKACCOUNT 257
#define TOKACCOUNTS 258
#define TOKACTION 259
#define TOKACTIONS 260
#define TOKADDHEADER 261
#define TOKADDTOCACHE 262
#define TOKAGE 263
#define TOKALL 264
#define TOKALLOWMANY 265
#define TOKAND 266
#define TOKANYNAME 267
#define TOKANYSIZE 268
#define TOKANYTYPE 269
#define TOKAPPEND 270
#define TOKATTACHMENT 271
#define TOKBODY 272
#define TOKBYTES 273
#define TOKCACHE 274
#define TOKCASE 275
#define TOKCMDUSER 276
#define TOKCOMPRESS 277
#define TOKCONTINUE 278
#define TOKCOUNT 279
#define TOKDAYS 280
#define TOKDEFUSER 281
#define TOKDELTOOBIG 282
#define TOKDISABLED 283
#define TOKDOMAIN 284
#define TOKDOTLOCK 285
#define TOKDROP 286
#define TOKEQ 287
#define TOKEXEC 288
#define TOKEXPIRE 289
#define TOKFCNTL 290
#define TOKFILEGROUP 291
#define TOKFILEUMASK 292
#define TOKFLOCK 293
#define TOKFOLDER 294
#define TOKFOLDERS 295
#define TOKFROM 296
#define TOKGIGABYTES 297
#define TOKGROUP 298
#define TOKGROUPS 299
#define TOKHEADER 300
#define TOKHEADERS 301
#define TOKHOURS 302
#define TOKIGNOREERRORS 303
#define TOKIMAP 304
#define TOKIMAPS 305
#define TOKIMPLACT 306
#define TOKIN 307
#define TOKINCACHE 308
#define TOKINSECURE 309
#define TOKINVALID 310
#define TOKKEEP 311
#define TOKKEY 312
#define TOKKILOBYTES 313
#define TOKLMTP 314
#define TOKLOCKFILE 315
#define TOKLOCKTIMEOUT 316
#define TOKLOCKTYPES 317
#define TOKLOCKWAIT 318
#define TOKLOOKUPORDER 319
#define TOKMAILDIR 320
#define TOKMAILDIRS 321
#define TOKMATCH 322
#define TOKMATCHED 323
#define TOKMAXSIZE 324
#define TOKMBOX 325
#define TOKMBOXES 326
#define TOKMEGABYTES 327
#define TOKMINUTES 328
#define TOKMONTHS 329
#define TOKNE 330
#define TOKNEWONLY 331
#define TOKNNTP 332
#define TOKNNTPS 333
#define TOKNOAPOP 334
#define TOKNOCRAMMD5 335
#define TOKNOCREATE 336
#define TOKNOLOGIN 337
#define TOKNONE 338
#define TOKNOPLAIN 339
#define TOKNORECEIVED 340
#define TOKNOT 341
#define TOKNOUIDL 342
#define TOKNOVERIFY 343
#define TOKXOAUTH2 344
#define TOKOAUTHBEARER 345
#define TOKOLDONLY 346
#define TOKOR 347
#define TOKPARALLELACCOUNTS 348
#define TOKPASS 349
#define TOKPASSWD 350
#define TOKPIPE 351
#define TOKPOP3 352
#define TOKPOP3S 353
#define TOKPORT 354
#define TOKPROXY 355
#define TOKPURGEAFTER 356
#define TOKQUEUEHIGH 357
#define TOKQUEUELOW 358
#define TOKREMOVEFROMCACHE 359
#define TOKREMOVEHEADER 360
#define TOKREMOVEHEADERS 361
#define TOKRETURNS 362
#define TOKREWRITE 363
#define TOKSECONDS 364
#define TOKSERVER 365
#define TOKSET 366
#define TOKSIZE 367
#define TOKSMTP 368
#define TOKSTARTTLS 369
#define TOKSTDIN 370
#define TOKSTDOUT 371
#define TOKSTRING 372
#define TOKSTRIPCHARACTERS 373
#define TOKTAG 374
#define TOKTAGGED 375
#define TOKTIMEOUT 376
#define TOKTO 377
#define TOKTOTALSIZE 378
#define TOKUNMATCHED 379
#define TOKUSER 380
#define TOKUSERS 381
#define TOKVALUE 382
#define TOKVERIFYCERTS 383
#define TOKWEEKS 384
#define TOKWRITE 385
#define TOKYEARS 386
#define NONE 387
#define NUMBER 388
#define STRING 389
#define STRMACRO 390
#define NUMMACRO 391
#define STRCOMMAND 392
#define NUMCOMMAND 393
#ifndef YYSTYPE_DEFINED
#define YYSTYPE_DEFINED
typedef union
{
	long long		 number;
	char			*string;
	int			 flag;
	u_int			 locks;
	struct {
		struct fetch	*fetch;
		void		*data;
	} fetch;
	struct {
		char		*host;
		char		*port;
	} server;
	enum area		 area;
	enum exprop		 exprop;
	struct actitem		*actitem;
	struct actlist		*actlist;
	struct expr		*expr;
	struct expritem		*expritem;
	struct strings		*strings;
	struct replstrs		*replstrs;
	enum fetch_only		 only;
	struct {
		char		*path;
		enum fetch_only	 only;
	} poponly;
	struct {
		int		 flags;
		char		*str;
	} re;
	gid_t			 localgid;
	enum cmp		 cmp;
	struct rule		*rule;
	struct {
		char		*user;
		int		 user_netrc;
		char		*pass;
		int		 pass_netrc;
	} userpass;
	userfunction		 ufn;
	struct userfunctions	*ufns;
} YYSTYPE;
#endif /* YYSTYPE_DEFINED */
extern YYSTYPE yylval;
